// ------------------------------------------------------------
// GICv3 example
//
// Copyright Arm Ltd 2017. All rights reserved.
// ------------------------------------------------------------

  .section  GIC,"ax"         // Define an executable ELF section
  .align 3                   // Align to 2^3 byte boundary

// ------------------------------------------------------------

// Defines used for GIC

.equ GICDbase,		          0x2F000000

.equ GICD_CTLRoffset,		  0x0
.equ GICD_CTLR.ARE_S,		  (1<<4)
.equ GICD_CTLR.ARE_NS,		  (1<<5)
.equ GICD_CTLR.EnableG0,      1
.equ GICD_CTLR.EnableG1NS,	  (1<<1)
.equ GICD_CTLR.EnableG1S,	  (1<<2)
.equ GICD_CTLR.DS,		      (1<<6)

.equ GICD_IGROUPRoffset,	  0x080
.equ GICD_ISENABLERoffset,    0x100
.equ GICD_ICENABLERoffset,    0x180
.equ GICD_ISPENDRoffset,      0x200
.equ GICD_ICPENDRoffset,      0x280
.equ GICD_IPRIORITYRoffset,   0x400
.equ GICD_ICFGRoffset,        0xC00
.equ GICD_IGRPMODRoffset,     0xD00
.equ GICD_NSACRoffset,        0xE00
.equ GICD_SGIR,               0xF00

.equ GICD_IROUTERoffset,      0x6000
.equ GICD_IROUTER.RM,         (1<<31)

// ------------------------------------------------------------

.equ ICC_SRE_ELn.Enable,     (1<<3)
.equ ICC_SRE_ELn.SRE,        (1)

// ------------------------------------------------------------

.equ RDbase,	               0x2F100000

.equ GICR_WAKERoffset,		      0x14
.equ GICR_WAKER.ProcessorSleep,  (1<<2)
.equ GICR_WAKER.ChildrenAsleep,  (1<<3)

.equ GICR_CTLRoffset,		     0x0

// ------------------------------------------------------------

.equ SGIbase,	               (RDbase + 0x10000)

.equ GICR_IGROUPRoffset,	 0x080
.equ GICR_ISENABLERoffset,   0x100
.equ GICR_ICENABLERoffset,   0x180
.equ GICR_ISPENDRoffset,     0x200
.equ GICR_ICPENDRoffset,     0x280
.equ GICR_IPRIORITYRoffset,  0x400
.equ GICR_ICFGRoffset,       0xC00
.equ GICR_IGRPMODRoffset,    0xD00
.equ GICR_NSACRoffset,       0xE00

// ------------------------------------------------------------
// Initialize GIC, and enable timer interrupt
// ------------------------------------------------------------

  .global gicInit
  .type gicInit, "function"
gicInit:
  // Configure Distributor
  MOV      x0, #GICDbase  // Address of GIC

  // Set ARE bits and group enables in the Distributor
  ADD      x1, x0, #GICD_CTLRoffset
  MOV      x2,     #GICD_CTLR.ARE_NS
  ORR      x2, x2, #GICD_CTLR.ARE_S
  STR      w2, [x1]

  ORR      x2, x2, #GICD_CTLR.EnableG0
  ORR      x2, x2, #GICD_CTLR.EnableG1S
  ORR      x2, x2, #GICD_CTLR.EnableG1NS
  STR      w2, [x1]
  DSB      SY

  // Configure Redistributor
  // Clearing ProcessorSleep signals core is awake
  MOV      x0, #RDbase
  MOV      x1, #GICR_WAKERoffset
  ADD      x1, x1, x0
  STR      wzr, [x1]
  DSB      SY
1:   // We now have to wait for ChildrenAsleep to read 0
  LDR      w0, [x1]
  AND      w0, w0, #0x6
  CBNZ     w0, 1b // NB 1b means branch to nearest label "1" BEFORE the current instruction

  // Configure CPU interface
  // We need to set the SRE bits for each EL to enable
  // access to the interrupt controller registers
  MOV      x0, #ICC_SRE_ELn.Enable
  ORR      x0, x0, #ICC_SRE_ELn.SRE
  MSR      ICC_SRE_EL3, x0
  ISB
  MSR      ICC_SRE_EL2, x0
  ISB
  MSR      ICC_SRE_EL1, x0
  
  // Set the secure version of ICC_SRE_EL1
  ISB
  MRS      x1, SCR_EL3         
  BIC	   w1, w1, #1          // Set NS bit (lower EL in secure state)
  MSR      SCR_EL3, x1
  
  ISB
  MSR      ICC_SRE_EL1, x0
  
  MRS      x1, SCR_EL3       
  ORR	   w1, w1, #1          // Set NS bit (lower EL in non secure state)
  MSR      SCR_EL3, x1
  
  ISB
  MOV      x0, #0xFF
  MSR      ICC_PMR_EL1, x0 // Set PMR to lowest priority
  
  ISB
  MOV      x0, #3
  MSR      ICC_IGRPEN1_EL3, x0
  
  ISB
  MOV      x0, #1
  MSR      ICC_IGRPEN1_EL1, x0
  MSR      ICC_IGRPEN0_EL1, x0
  ISB

  MOV      x0, #SGIbase       // Address of Redistributor registers

  ADD      x1, x0, #GICR_IGROUPRoffset
  MOV      w3, #0xffffffff
  STR      w3, [x1]          // Mark INTIDs 0..31 as Non Secure

  ADD      x1, x0, #GICR_IGRPMODRoffset
  STR      wzr, [x1]          // Mark INTIDs 0..31 as Non Secure Group 1

  ADD      x1, x0, #GICR_ISENABLERoffset
  MOV      w2, #(1 << 30)     // Enable INTID 30
  STR      w2, [x1]           // Enable interrupt source
  
  ADD      x1, x0, #GICR_NSACRoffset
  MOV      w2, #(1 << 30)     // Enable INTID 30
  STR      w2, [x1]           // Enable interrupt source
  
  ISB
  RET

// ------------------------------------------------------------

  .global readIAR1
  .type readIAR1, "function"
readIAR1:
  MRS       x0, ICC_IAR1_EL1  // Read ICC_IAR1_EL1 into x0
  RET

// ------------------------------------------------------------

  .global writeEOIR1
  .type writeEOIR1, "function"
writeEOIR1:
  MSR        ICC_EOIR1_EL1, x0 // Write x0 to ICC_EOIR0_EL1
  RET

// ------------------------------------------------------------
// End of file
// ------------------------------------------------------------
