// ------------------------------------------------------------
// ARMv8-A AArch64 Startup Example
//
// Copyright Arm Ltd 2017. All rights reserved.
// ------------------------------------------------------------

  .section  BOOT,"ax"        // Define an executable ELF section, BOOT
  .align 3                   // Align to 2^3 byte boundary
 
// ------------------------------------------------------------
.equ AArch64_EL2_SP2, 0x09 // EL2h
.equ AArch64_EL2_SP0, 0x08 // EL2t
.equ AArch64_EL1_SP1, 0x05 // EL1h
.equ AArch64_EL1_SP0, 0x04 // EL1t
.equ AArch64_EL0_SP0, 0x00

// ------------------------------------------------------------
// EL3 AArch64
// ------------------------------------------------------------
  
  .global start64
  .type start64, "function" 
start64:
  // Enter EL3, configure system registers to enable switching between secure exceptions levels

  // Which core am I
  // ---------------
  MRS      x0, MPIDR_EL1
  AND      x0, x0, #0xFFFF                   // Mask off to leave Aff0 and Aff1
  CBZ      x0, boot                          // If not *.*.0.0, then go to sleep
sleep:
  WFI
  B        sleep
boot:
  ADRP x0, Image$$STACK_EL3$$ZI$$Limit
  MOV  sp, x0
  

  // Disable trapping of CPTR_EL3 accesses or use of Adv.SIMD/FPU
  // ----------------------------------------------------------
  MOV      x0, #0                           // Clear all trap bits
  MSR      CPTR_EL3, x0

  
  // Configure SCR_EL3
  // -----------------
  MOV      w1, #0              // Initial value of register is unknown
  ORR 	   w1, w1, #(1 << 11)  // set ST bit (disable trapping of timer control registers)
  ORR	   w1, w1, #(1 << 10)  // set RW bit (next lower EL in aarch64)
  ORR      w1, w1, #(1 << 3)   // Set EA bit (SError routed to EL3)
  ORR      w1, w1, #(1 << 2)   // Set FIQ bit (FIQs routed to EL3)
  ORR      w1, w1, #(1 << 1)   // Set IRQ bit (IRQs routed to EL3)
  MSR      SCR_EL3, x1


  // Install vector table
  // --------------------
  .global vectors
  LDR  x0, =vectors
  MSR  VBAR_EL3, x0
  
  // Clear interrupt masks
  // ----------------------
  MSR      DAIFClr, #0xF

  // Initialize SCTLR_EL1
  // --------------------
  // SCTLR_EL1 has an unknown reset value and must be configured
  // before we can enter EL1
  MSR 	   SCTLR_EL1, xzr
  
  LDR      x0, =el1_entry_aarch64
  LDR	   x1, =AArch64_EL1_SP1
  MSR 	   ELR_EL3, x0 // where to branch to when exception completes
  MSR      SPSR_EL3, x1 // set the program state for this point to a known value
  
  BL       gicInit
  
  ISB
  ERET

// ------------------------------------------------------------
// EL1 AArch64
// ------------------------------------------------------------
    .global el1_entry_aarch64
	.type el1_entry_aarch64, "function"
	
el1_entry_aarch64:
  // we can use the same vector table in this example, but in general
  // each combination of exception level, security state, and execution state
  // will need a new vector table
  LDR      x0, =vectors
  MSR	   VBAR_EL1, x0
	
  //we must ensure that floating point register accesses are not trapped
  //since the c library for AArch64-v8A uses them
  MOV      x0, #(0x3 << 20)
  MSR      CPACR_EL1, x0
  
  // ensures that all instructions complete before 
  ISB 
  // Branch to scatter loading and C library init code
  .global  __main
  B        __main

// ------------------------------------------------------------
// End of file
// ------------------------------------------------------------
